# TODO: remove dependency to streamlitfront

# from streamlitfront import dispatch_funcs
# from front.scrap.pydantic_wrap import SimplePageFuncPydanticWithOutput
#
#
# def identity(x:int)->int:
# 	return x
#
# funcs = [identity]
#
# configs = {'page_factory': SimplePageFuncPydanticWithOutput}
#
#
# app = dispatch_funcs(funcs, configs=configs)
# app()
